@@ -43,18 +43,31 @@ in this Software without prior written authorization from The Open Group.
  * THIS SOFTWARE.
  */
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include <X11/fonts/FSlib.h>
 #include <stdio.h>
 #include <X11/Xos.h>
 #include <stdlib.h>
+#include <assert.h>
+
+#ifdef HAVE_BSD_STDLIB_H
+#include <bsd/stdlib.h>
+#endif
+
+#ifndef HAVE_REALLOCARRAY
+#define reallocarray(old, num, size) realloc(old, (num) * (size))
+#endif
 
 #ifndef N_START
 #define N_START 1000		/* Maximum # of fonts to start with */
 #endif
 
-static int max_output_line_width = 79;
-static int output_line_padding = 3;
-static int columns = 0;
+static unsigned int max_output_line_width = 79;
+static unsigned int output_line_padding = 3;
+static unsigned int columns = 0;
 
 #define L_SHORT 0
 #define L_MEDIUM 1
@@ -64,7 +77,7 @@ static int columns = 0;
 static Bool sort_output = True;
 static int  long_list = L_SHORT;
 static int  nnames = N_START;
-static int  font_cnt;
+static unsigned int  font_cnt;
 static int  min_max;
 typedef struct {
     char       *name;
@@ -79,18 +92,30 @@ static FSServer *svr;
 
 static char *program_name;
 
-static void usage ( void ) _X_NORETURN;
+static void usage (const char *msg) _X_NORETURN _X_COLD;
 static void get_list ( const char *pattern );
 static int compare ( const void *f1, const void *f2 );
 static void show_fonts ( void );
 static void print_font_header ( void );
 static void show_font_header ( FontList *list );
-static void copy_number ( char **pp1, char **pp2, int n1, int n2 );
+static void copy_number ( char **pp1, char **pp2, char **ep1, char **ep2,
+                          int n1, int n2, char suffix );
 static void show_font_props ( FontList *list );
 
-static void
-usage(void)
+static void _X_NORETURN _X_COLD
+missing_arg (const char *option)
 {
+    char msg[32];
+
+    snprintf(msg, sizeof(msg), "%s requires an argument", option);
+    usage(msg);
+}
+
+static void
+usage(const char *msg)
+{
+    if (msg)
+	fprintf(stderr, "%s: %s\n", program_name, msg);
     fprintf(stderr, "usage:  %s [-options] [-fn pattern]\n", program_name);
     fprintf(stderr, "%s", "where options include:\n"
 	    "    -l[l[l]]                 give long info about each font\n"
@@ -101,6 +126,7 @@ usage(void)
 	    "    -w width                 maximum width for multiple columns\n"
 	    "    -n columns               number of columns if multi column\n"
 	    "    -server servername       font server to contact\n"
+	    "    -version                 print command version and exit\n"
 	    "\n");
     exit(1);
 }
@@ -117,15 +143,18 @@ main(int argc, char *argv[])
     for (i = 1; i < argc; i++) {
 	if (strncmp(argv[i], "-s", 2) == 0) {
 	    if (++i >= argc)
-		usage();
+		missing_arg("-server");
 	    servername = argv[i];
 	}
+	else if (strcmp(argv[i], "-version") == 0) {
+	    printf("%s\n", PACKAGE_STRING);
+	    exit(0);
+	}
     }
 
     if ((svr = FSOpenServer(servername)) == NULL) {
 	if (FSServerName(servername) == NULL) {
-	    fprintf(stderr, "%s: no font server defined\n", program_name);
-	    exit(0);
+	    usage("no font server defined");
 	}
 	fprintf(stderr, "%s:  unable to open server \"%s\"\n",
 		program_name, FSServerName(servername));
@@ -135,7 +164,7 @@ main(int argc, char *argv[])
     for (argv++, argc--; argc; argv++, argc--) {
 	if (argv[0][0] == '-') {
 	    if (argcnt > 0)
-		usage();
+		usage(NULL);
 	    for (i = 1; argv[0][i]; i++)
 		switch (argv[0][i]) {
 		case 'l':
@@ -152,37 +181,38 @@ main(int argc, char *argv[])
 		    break;
 		case 'f':
 		    if (--argc <= 0)
-			usage();
+			missing_arg("-fn");
 		    argcnt++;
 		    argv++;
 		    get_list(argv[0]);
 		    goto next;
 		case 'w':
 		    if (--argc <= 0)
-			usage();
+			missing_arg("-w");
 		    argv++;
-		    max_output_line_width = atoi(argv[0]);
+		    max_output_line_width = (unsigned int) atoi(argv[0]);
 		    goto next;
 		case 'n':
 		    if (--argc <= 0)
-			usage();
+			missing_arg("-n");
 		    argv++;
-		    columns = atoi(argv[0]);
+		    columns = (unsigned int) atoi(argv[0]);
 		    goto next;
 		case 'u':
 		    sort_output = False;
 		    break;
 		case 's':	/* eat -s */
 		    if (--argc <= 0)
-			usage();
+			missing_arg("-server");
 		    argv++;
 		    goto next;
 		default:
-		    usage();
-		    break;
+		    fprintf(stderr, "%s: unrecognized option '%s'\n",
+			    program_name, argv[0]);
+		    usage(NULL);
 		}
 	    if (i == 1)
-		usage();
+		usage(NULL);
 	} else {
 	    argcnt++;
 	    get_list(argv[0]);
@@ -242,12 +272,20 @@ get_list(const char *pattern)
 		program_name, pattern);
 	return;
     }
-    if (font_list)
-	font_list = (FontList *) realloc(font_list,
-				  (font_cnt + available) * sizeof(FontList));
-    else
-	font_list = (FontList *) malloc((unsigned)
-				  (font_cnt + available) * sizeof(FontList));
+    else {
+	FontList *old_list = font_list;
+
+	font_list = reallocarray(old_list, (font_cnt + (unsigned) available),
+                                 sizeof(FontList));
+
+	if (font_list == NULL) {
+	    free(old_list);
+	    fprintf(stderr, "%s: unable to allocate %zu bytes for font list\n",
+		    program_name,
+		    (font_cnt + (unsigned) available) * sizeof(FontList));
+	    exit(-1);
+	}
+    }
     for (i = 0; i < available; i++) {
 	font_list[font_cnt].name = fonts[i];
 
@@ -276,7 +314,7 @@ compare(const void *f1, const void *f2)
 static void
 show_fonts(void)
 {
-    int         i;
+    unsigned int i;
 
     if (font_cnt == 0)
 	return;
@@ -303,14 +341,14 @@ show_fonts(void)
 	return;
     }
     if ((columns == 0 && isatty(1)) || columns > 1) {
-	int         width,
+	unsigned int width,
 	            max_width = 0,
 	            lines_per_column,
 	            j,
 	            index;
 
 	for (i = 0; i < font_cnt; i++) {
-	    width = strlen(font_list[i].name);
+	    width = (unsigned int) strlen(font_list[i].name);
 	    if (width > max_width)
 		max_width = width;
 	}
@@ -409,35 +447,26 @@ show_font_header(FontList *list)
 	            max[BUFSIZ];
 	char       *pmax = max,
 	           *pmin = min;
+	char       *emin = min + sizeof(min),
+	           *emax = max + sizeof(max);
 
-	strcpy(pmin, "     min(l,r,w,a,d) = (");
-	strcpy(pmax, "     max(l,r,w,a,d) = (");
-	pmin += strlen(pmin);
-	pmax += strlen(pmax);
-
-	copy_number(&pmin, &pmax,
+	copy_number(&pmin, &pmax, &emin, &emax,
 		    pfh->min_bounds.left,
-		    pfh->max_bounds.left);
-	*pmin++ = *pmax++ = ',';
-	copy_number(&pmin, &pmax,
+		    pfh->max_bounds.left, ',');
+	copy_number(&pmin, &pmax, &emin, &emax,
 		    pfh->min_bounds.right,
-		    pfh->max_bounds.right);
-	*pmin++ = *pmax++ = ',';
-	copy_number(&pmin, &pmax,
+		    pfh->max_bounds.right, ',');
+	copy_number(&pmin, &pmax, &emin, &emax,
 		    pfh->min_bounds.width,
-		    pfh->max_bounds.width);
-	*pmin++ = *pmax++ = ',';
-	copy_number(&pmin, &pmax,
+		    pfh->max_bounds.width, ',');
+	copy_number(&pmin, &pmax, &emin, &emax,
 		    pfh->min_bounds.ascent,
-		    pfh->max_bounds.ascent);
-	*pmin++ = *pmax++ = ',';
-	copy_number(&pmin, &pmax,
+		    pfh->max_bounds.ascent, ',');
+	copy_number(&pmin, &pmax, &emin, &emax,
 		    pfh->min_bounds.descent,
-		    pfh->max_bounds.descent);
-	*pmin++ = *pmax++ = ')';
-	*pmin = *pmax = '\0';
-	printf("%s\n", min);
-	printf("%s\n", max);
+		    pfh->max_bounds.descent, '\0');
+	printf("     min(l,r,w,a,d) = (%s)\n", min);
+	printf("     max(l,r,w,a,d) = (%s)\n", max);
     }
 }
 
@@ -445,44 +474,49 @@ show_font_header(FontList *list)
 #define	max(a, b)	((a) > (b) ? (a) : (b))
 #endif
 
+/*
+ * Append string representations of n1 to pp1 and n2 to pp2,
+ * followed by the given suffix character,
+ * but not writing into or past ep1 & ep2, respectively.
+ * The string representations will be padded to the same width.
+ */
 static void
-copy_number(char **pp1, char **pp2, int n1, int n2)
+copy_number(char **pp1, char **pp2, char **ep1, char **ep2, int n1, int n2,
+    char suffix)
 {
     char       *p1 = *pp1;
     char       *p2 = *pp2;
-    int         w;
+    int         w, w1, w2;
 
-    sprintf(p1, "%d", n1);
-    sprintf(p2, "%d", n2);
-    w = max(strlen(p1), strlen(p2));
-    sprintf(p1, "%*d", w, n1);
-    sprintf(p2, "%*d", w, n2);
-    p1 += strlen(p1);
-    p2 += strlen(p2);
-    *pp1 = p1;
-    *pp2 = p2;
+    w1 = snprintf(NULL, 0, "%d", n1);
+    w2 = snprintf(NULL, 0, "%d", n2);
+    w = (int) max(w1, w2);
+    assert(w > 0);
+    snprintf(p1, *ep1 - p1, "%*d%c", w, n1, suffix);
+    snprintf(p2, *ep2 - p2, "%*d%c", w, n2, suffix);
+    *pp1 = p1 + strlen(p1);
+    assert(*pp1 < *ep1);
+    *pp2 = p2 + strlen(p2);
+    assert(*pp2 < *ep2);
 }
 
 static void
 show_font_props(FontList *list)
 {
-    int         i;
-    char        buf[1000];
+    unsigned int  i;
     FSPropInfo *pi = list->pi;
     FSPropOffset *po = list->po;
     unsigned char *pd = list->pd;
-    int         num_props;
+    unsigned int  num_props;
 
     num_props = pi->num_offsets;
     for (i = 0; i < num_props; i++, po++) {
-	strncpy(buf, (char *) (pd + po->name.position), po->name.length);
-	buf[po->name.length] = '\0';
-	printf("%s\t", buf);
+	fwrite(pd + po->name.position, 1, po->name.length, stdout);
+	putc('\t', stdout);
 	switch (po->type) {
 	case PropTypeString:
-	    strncpy(buf, (char *)pd + po->value.position, po->value.length);
-	    buf[po->value.length] = '\0';
-	    printf("%s\n", buf);
+	    fwrite(pd + po->value.position, 1, po->value.length, stdout);
+	    putc('\n', stdout);
 	    break;
 	case PropTypeUnsigned:
 	    printf("%lu\n", (unsigned long) po->value.position);
